// ------------------------------------------------------------
// File       : webparse.cpp
// Created    : September 26, 2003
// Author     : M. E. Shmelyov
// Description: Web form submitted data parser
// Instruction: Call ParseForm() function before any
//              operations with form data. Only content-type
//              application/x-www-form-urlencoded is accepted.
//              If form was submitted with content-type
//              multipart/form-data function does not process
//              form data and returns false.
//              On success ParseForm() returns true.
//              After this all form data will be contained
//              in formdata variable. It is associative
//              array (map). You can get any control-associated
//              value by using subscript.
//              E. g. formdata["user"] gives you string with
//              value of "user" form element if such exist.
//              <OPTION> values from <SELECT> that allows
//              multiple selection are stored under same key
//              and are separated by semicolons ";".
// ------------------------------------------------------------

#include "webparse.h"
#include <stdlib.h>
#include <stdio.h>

using namespace std;

map<string, string> formdata;


bool ParseForm(bool ignore_cr)
{
  bool IsGET, Doing=true, TheKey=true, AddChar;
  string contentType, formMethod;
#define Key contentType
#define Value formMethod
  char *ContentType, *FormMethod, *QueryString, *qs;
  unsigned char ch1, ch2;
  int intch1, escStep=0, firstH, secondH;
#define hexNum intch1
  ContentType = getenv("CONTENT_TYPE");
  if (ContentType!=NULL) contentType = ContentType;
  if (contentType == "multipart/form-data")
    return false;
  FormMethod = getenv("REQUEST_METHOD");
  if (FormMethod!=NULL) formMethod = FormMethod;
  IsGET = (formMethod == "GET");
  //if not GET then POST is assumed
  if (IsGET)
  {
    QueryString = getenv("QUERY_STRING");
    if (QueryString==NULL) return true;
    qs = QueryString;
  }
  //printf("debug: before Key.resize(0);  Value.resize(0);\n");
  Key.resize(0);  Value.resize(0);
  //printf("debug: before while\n");
  while (Doing)
  {
    //printf("debug: while step\n");
    if (IsGET)
    {//GET
      ch1 = (*qs);  qs++;
      if (ch1==0)
        Doing = false;
    }
    else
    {//POST
      intch1 = getc(stdin);
      if (intch1==EOF)
        Doing = false;
      else
        ch1 = (unsigned char)intch1;
    }
    if (Doing)
    {
      AddChar = true;
      if (escStep==0)
      {
        switch (ch1)
        {
        case '%': escStep = 1; break;
        case '+': ch2 = ' '; break;
        case '=': TheKey=false;  AddChar = false; break;
        case '&':
         if (formdata.count(Key)>0)
          { formdata[Key] += ';';  formdata[Key] += Value; }
         else
          formdata[Key] = Value;
         Key.resize(0);  Value.resize(0);  TheKey = true;
         AddChar = false;
        break;
        default: ch2 = ch1;
        }
      }
      else
      {//parsing escape sequence
        if (escStep==1)
        {
          if (ch1=='%')
          {
            ch2 = '%';  escStep = 0;
          }
          else
          {
           firstH = ch1;
           if (firstH<58) firstH -= 48; //digits
           else
           {
            if (firstH<91) firstH -= 55; //capital Latin
            else firstH -= 87; //small Latin
           }
           escStep++;
          }
        }
        else
        {
          secondH = ch1;
          if (secondH<58) secondH -= 48; //digits
          else
          {
            if (secondH<91) secondH -= 55; //capital Latin
            else secondH -= 87; //small Latin
          }
          hexNum = (firstH<<4) + secondH;
          ch2 = (unsigned char)hexNum;
          escStep = 0;
          if (ignore_cr && (ch2=='\r'))
            AddChar = false;
        }
      }//end else (escStep==0)
      AddChar = AddChar && (escStep==0);
      if (AddChar)
      {
        if (TheKey) Key += ch2;
        else Value += ch2;
      }
    }//end if (Doing)
    else
    {
      if (!Key.empty())
      {
         if (formdata.count(Key)>0)
          { formdata[Key] += ';';  formdata[Key] += Value; }
         else
          formdata[Key] = Value;
      }
    }
  }//end while
  return true;
}//end ParseForm